/*
  h_file.c   Nov98  JS  version 0.3
  
  this is a file for handling Htmp-File (tmp_hxy.000) 
  reasons and goals for this file:
  - 2GB file breaking => use LFS (64bit) or more blocks
  - error handling (perror nach printf???) Jun99
  - open/close
  - hidden variables (encapsulating of file-handling)
  - better reading of main-file
  Ideas:
  - chksum for tmp-file, check during read?
  - vvv==1 ? measure disk performance
  - compressed tmp_hxy.XXX ? (example: 1GB => 336MB, zip=17m  unzip=3m read=1m)
                        thor  gzip -c1 1GB => 640MB, zip=10m  unzip=3m
                        thor  gzip -c9 1GB => 579MB, zip=125m unzip=3m 
                        thor  gzip -c4 1GB => 550MB, zip=14m  unzip=3m
  - see pipe.c!!! (h_file,fopen,fclose,... & fn,pipe_ropen,... kapseln
  - repeat H*v if error!?
  
  this file is still changing Nov98,Apr99,Jun99,Nov99
  history:
   v2.17 - array of thxy -to- thxy of arrays
   v2.0.1 - access to mcplx only via Macros M_d2f(), M_f2d()
   v2.0.0 - file-split removed (see history)
   03.02.00 - bug cplx rr => mcplx rr
   25.11.99 - bug in my_fread/write if file > 1MB beseitigt
   07.01.00 - bug auf MIPS beseitigt (high-low Format!!!)
               Vorsicht bei type-Umwandlungen (high-low/low-high format)

*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>	/* strerror */
#include "defs.h"
#include "spins.h"
#include "h_file.h"
#include "pipe.h"
#include "block.h"
#if 1
#include <unistd.h>		/* test this with configure! */
#define SLEEP(x) sleep(x)
#else
#define SLEEP(x) {}
#endif

/*
 * v2.30 can have different numbers of I/O threads and compute threads
 */

#ifndef HRMAX
#if HRMAX>0
#error "HRMAX!=0 if not defined?" 
#endif
#endif
#if HRMAX>0	/* table for 16bit index to Hamilton-entries */
 unsigned int hr_len[B_NUM]; /* one table per buffer (thread) */
 cplx hr[B_NUM][HRMAX];	/* better use realloc for dyn. B_NUM ? */
#endif

/* define the basename for the H-files htmpBBB.PPP
 * where BBB is the IO-thread-number (block)
 * and PPP the paramter index (ToDo)
 */
#define Htmp "./tmp/htmp"	/* ./tmp can be a link to /tmp/$USER */

/* public variables */

/* privat variables */
char  h_name[F_NUM][40]; /* file-name F_NUM=(ToDo: <=)B_NUM files */
int   h_error=0;         /* set if H could not be written */
int   h_state[B_NUM];    /* 0=closed, 1=ropen, 2=wopen */
int   h_open[F_NUM];    /* number cthreads have files opened 0..B_NUM/F_NUM-1 */

/* my comfortable erroroutput for misterious errors */
int my_ferror(char *msg){
  perror(msg);	/* errno_msg => stderr */
  SLEEP(1);
  return 0;
}

/* we have F_NUM files for B_NUM compute threads,
 * ToDo: F_NUM async I/O tasks
 * because random access to disk is slow F_NUM should be less than B_NUM
 * but because programming it needs time and we finaly want to
 * remove H files from the code (if computation is fast enough)
 * we live with B_NUM=F_NUM
 */
#if B_NUM != F_NUM
#error "F_NUM!=B_NUM not implemented for pthreads, and sensless for single thread"
#endif

/* including error-handling (public???) */
size_t my_fread( char *ptr, size_t size, size_t nmemb, int block){
  size_t n_readed=0,nf,ni,tries=0; /* sgi-CC can not add void* */
  nmemb*=size;
  if (block>=F_NUM) ERR(9999);
  if( nmemb )
  while(n_readed< size /* nmemb */ ){ /* mind. 1 element */
    /* some clibs do alloc a buffer for disk_read which cost time if it is to big */
    ni=nmemb-n_readed; if(ni>262144) ni=262144;	/* 2^n=256kB pieces */
    nf=pipe_read(ptr,(int)ni,block);
    if(nf<=0){ E3("read=%d ni=%d block=%d ",(int)nf,(int)ni,block);
               my_ferror("read");if(tries++<5)continue;break; }
    n_readed+=nf; ptr+=nf; tries=0;
  } /* while */
  return n_readed/size;
}

void error999(){}
/* 
   write large objects in small 1e6-pieces,
    because fwrite (probably) need lot of memory
 */
size_t my_fwrite( void *ptr, size_t size, size_t nmemb, int block){
  size_t n_written=0,nf,ni,tries=0;char *p=(char *)ptr;	/* sgi-CC can not add void* */
  if (vvv&16)
    printf("\nvvv&16 blk %2d fwrite %d*%d %p err=%d",
           block, (int)size, (int)nmemb, ptr, h_error);
  if (block>=F_NUM) ERR(9999); /* cant be! */
  nmemb*=size;
  if (h_error) return 0;  /* dont try if something went wrong on I/O */
  while(n_written<nmemb){
    ni=nmemb-n_written; if(ni>262144) ni=262144;	/* 256kB pieces */
    nf=pipe_write(p,(int)ni,block);
    n_written+=nf; p+=nf;
    if(nf<ni){ ERR(620); error999();
      if (h_error) break;
      if (tries++<5) continue;
      break; }
    tries=0;
  } /* end while */
  /* if (vvv&16) printf(" %d/%d -",n_written,size); */
  return n_written/size;
}

/* --- end of file handler --- */

/* --------- interface between compute blocks and H storage ------ */

/* open for reading */
int h_ropen(int block){ /* 0..B_NUM-1, block > F_NUM possible */
  if (vvv&16) printf("\nvvv&16 blk %2d h_ropen", block);
  if (block<0 || block>=B_NUM) ERR(9999); /* for safety */
  if (++h_open[block%F_NUM]==1) { /* open stream for first block */
    if (vvv&16) printf(" popen %s", h_name[block%F_NUM] );
    if ( pipe_ropen(h_name[block],block) ){
      E1("fopen_%s",h_name[block]);return 1; }
  }
  b_hbl[block]=HBLen; /* force h_read on first h_get */
  if (h_state[block]!=0) ERR(9999); /* file open */
  h_state[block]=1;
  return 0;
}

/* open for writing */
int h_wopen(int block){  /* 0..B_NUM-1, block > F_NUM possible */
  if (block<0 || block>=B_NUM) ERR(9999); /* for safety */
  if(vvv&1){
    if(block==0) printf(" write " Htmp "*");
    if (!(vvv&16)) printf(" %d",block); /* minimum info */
  }
    if (vvv&16) /* maximum info */
      printf("\nvvv&16 blk %2d h_wopen open=%d", block, h_open[block%F_NUM]);
  if (++h_open[block%F_NUM]==1) { /* open stream for first block */
    if (vvv&16)
      printf(" pipe_wopen %2d", block%F_NUM);
    if ( pipe_wopen(h_name[block%F_NUM],block%F_NUM) ) { 
      E1("wopen %s",h_name[block%F_NUM]); h_error=1; return 1; }
  }
  b_hbl[block]=0;
  if (h_state[block]!=0) ERR(9999); /* file not open */
  h_state[block]=2;
  return 0;
}

static int out_vvv=0;

int h_rclose(int block) {
  if (h_state[block]!=1) ERR(9999); /* file not open */
  h_state[block]=0;
  if (--h_open[block%F_NUM]==0) /* close stream if last block flushed */
    pipe_close(block%F_NUM);
  return 0;
}

/* --- read from Htmp-file, ret: 0=ok --- 
 * read from disk into H-buffer, struct thxy = xy[HBLen], r[HBLen](or idx+para)
 *  ToDo: get special block from stream of random(?) stored blocks 
 *  search dynamic buffer, if not found extend buffer until block found
 *  max. B_NUM*B_NUM/F_NUM (* PARAM)?
 * this buffer is written by F_NUM IO-threads
 *   and read by B_NUM compute threads (v2.26 deadlocks?)
 * IO-thread=file blocks=compute-threads
 *                          v-blocks IO=0-Buffer? ToDo
 *    0:          0,1,4,0,5,0,1,4 (compute block 0 was faster than the others?)
 *    1:          2,3,6,7,2,3,6,7
 */
/* ToDo: static struct thxy hblockbufs[F_NUM][B_NUM]; */
/* ToDo: need only B_NUM/F_NUM per IO-thread? */
 
/* privat */
/* read special block, !!! ToDo: read next block, caller must handle that */
/* ToDo: read and fill one of the freed buffer, better use FiFo 
 *  and let the caller work on every block coming from IO-stream
 * its slow but hopefully work for nonparallel, replace later by better 
 */ 
int h_read(int block){
  /* int ft=block%F_NUM; */ /* num of responsible IO-thread not used */
  /* int i; */
  if (vvv&16 && out_vvv<1000) printf("\nvvv&16 blk %2d h_read",block);
  if (h_state[block]!=1) ERR(9999); /* file not open */
  if (my_fread((char *)&b_hbuf[block],Hsize,1,block%F_NUM)!=1){
     E0("h_read"); ERR(9622); return 1;
   }
  if (vvv&16 && out_vvv<1000) printf(" ask=%2d got=%3d", block, b_hbuf[block].block);
  if (vvv&16)  /* move to h_read */
    printf("\nvvv&16 blk %2d h_read blk=%2d state=%d x= %6ld y= %6ld r=%3d ...",
          block, b_hbuf[block].block, h_state[block],
          (long) b_hbuf[block].xx,
          (long) b_hbuf[block].xy[ 0 ],
           (int) b_hbuf[block].r[ 0 ]);
  if (b_hbuf[block].block!=block) ERR(9999); /* blocks mixed */
  return 0;
}


/* ------------------------------------------------------------------ *
 * h_put: write matrix-element and its column (line encoded)          *
 * h_get: read  matrix-element and its column (line+1 if column>n1)   *
 * ------------------------------------------------------------------ */
#define HR_SIZE   sizeof(HR_TYPE)
#define XY_SIZE   sizeof(XY_TYPE)  


#if HRMAX>0
/* seperate error routine, only loaded to CPU-cache in error-case */
void err_fulltable(){
    int i; unsigned int j;
    /* output maximum on information, what was went wrong? */
    printf("\nerr_fulltable: --debug--\n");
    for (i=0;i<B_NUM*B_NUM;i++) {
      printf(" Hblock=[%2d,%2d] b_hnz=%9ld hr_len=%4d hr[]=",
        i/B_NUM,i%B_NUM,b_hnz[i],hr_len[i]);
      for (j=0;j<hr_len[i] && j<100;j++) printf(" % e",D(hr[i][j]));
      if (j>=100) printf(" ...");
      printf("\n");
    }
    ERR(9100);
}
#endif

/* h_put fills a write buffer, this fkt writes the buffer out!
 * h_wclose is only called to close the file after writing
 * block=1..B_NUM-1 can be equal and greater than F_NUM to flush all buffers
 */
int h_wclose(int block){  /* store in file */
  if(h_error) return 1;
  if (vvv&16)
    printf("\nvvv&16 blk %2d h_wclose open=%d", block, h_open[block%F_NUM]);
  if(b_hbl[block]){ /* buffer filled? */
   int i; /* fill rest of buffer, just to be sure and better debugging */
   for (i=b_hbl[block];i<HBLen;i++){
     b_hbuf[block].xy[i]=~0;
     b_hbuf[block].r[i]=0;
   }
   // printf(" hwrite %d (close)",block%F_NUM);
   if (h_state[block]!=2) ERR(9999); /* file not open */
   b_hbuf[block].block=block;
   if (vvv&16)
     printf("\nvvv&16 blk %2d h_write.block=%2d", block, b_hbuf[block].block);
   if( my_fwrite((char *)&b_hbuf[block], Hsize, 1, block%F_NUM) != 1 ) {
     h_error=1; return 1;
   }
  }
  if (h_state[block]!=2) ERR(9999); /* file not open */
  b_hbl[block]=0;
  h_state[block]=0; /* mark as closed */
  if (--h_open[block%F_NUM]==0) { /* close stream if last block flushed */
    if (vvv&16)
      printf(" pipe_close %2d", block%F_NUM);
    pipe_close(block%F_NUM); }
  return 0;
}

/* h_put and h_get moved to spins.c
 * to make the call shorter and allow inlining => no speedup Jan07 */

void ini_thxy(ULong n1){
  ULong i1,i2,i;int xy,j;struct thxy t1;
  for(i=0;i<B_NUM;i++){
    h_state[i]=0; /* default 0=close */
#if HRMAX>0
    hr_len[i]=0;
#endif
  }
  for(i1=0;i1<F_NUM;i1++){
    h_open[i1]=0;               /* no bocks open */
    h_name[i1][0]=0;		/* file-name ToDo: set name here! */
    sprintf(h_name[i1],Htmp "%03u.%03u",(int)i1,1); /* ToDo: set by user? per disk */
/*    for (i2=0;i2<B_NUM;i2++)
      hblockbufs[i1][i2].block=-1;
*/ /* mark as free, ToDo */
  }
  /* measure: i1 = used bytes for xy + XY_FLAG,
   *          i2 = sizeof xy in bytes (ToDo: check sense)
   * ToDo: since v2.19 we store y-b_ofs[blk] (not y), less bits
   */
  for(i1=0,i=2*n1;i!=0;i1++,i>>=8);
  for(i2=0,xy=1;xy!=0;i2++,xy<<=8);
  if(vvv)printf(" bytes(thxy.xy) %ld bytes(n1) %ld",(long)i2,(long)i1);
  if(  i1> i2)F0("xy to small, change XY_TYPE in h_file.h");
  if(2*i1<=i2)printf("\nWarning: change XY_TYPE in h_file.h to spare memory.");
  j=sizeof(int); /* compiler warnings for if(const1==const2) */
  if(j!=sizeof(t1.xy[0]))E0("sizeof int!=thxz.xy");
}

#define HR_FILE "tmp/hr_save.dat"

void close_thxy(int vvv){
  unsigned int i,hrmax;
  FILE *f1=NULL;
  /* if( save_mode & 1 ) --- small amount of data! save every time */
  { /* save data before possible crash */
    f1=fopen(HR_FILE,"wb");if(!f1)E0("error fopen"); }
  for(hrmax=i=0;i<B_NUM;i++){
#if HRMAX>0
    if( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if( f1 ){
      if(vvv)
        printf(((i==0)?" save hr[] len: %d":" %d"),hr_len[i]);
      fwrite(&hr_len[i],sizeof(hr_len[i]), 1, f1);
      fwrite(&hr[i][0], sizeof(hr[i][0]), hr_len[i], f1); /* hr_len==0 ??? */
    }
#endif
    fwrite(b_hnz, sizeof(b_hnz[0]), B_NUM, f1);
    fwrite(b_snz, sizeof(b_snz[0]), B_NUM, f1);
    fwrite(b_sn1, sizeof(b_sn1[0]), B_NUM, f1);
  }
  if(f1)fclose(f1);
  if(vvv){ printf(" close_thxy: max hr_num= %5d (entries/block)\n",hrmax); }
}

void restore_thxy(int vvv){
  unsigned int i,hrmax;
  FILE *f1=NULL;
  /* load data after crash */
  f1=fopen(HR_FILE,"rb");if(!f1)E0("error fopen");
#if HRMAX>0
  for(hrmax=i=0;i<B_NUM;i++){
    if(vvv)printf(" load hr[%d]=",i);
    fread(&hr_len[i],sizeof(hr_len[i]),1,f1);
    fread(&hr[i][0], sizeof(hr[i][0]), hr_len[i],f1);
    if( hr_len[i]>hrmax ) hrmax=hr_len[i];
    if(vvv)printf("%d",hr_len[i]);
  }
#endif
  fread(b_hnz, sizeof(b_hnz[0]), B_NUM, f1);
  fread(b_snz, sizeof(b_snz[0]), B_NUM, f1);
  fread(b_sn1, sizeof(b_sn1[0]), B_NUM, f1);
  if(f1)fclose(f1);
  if(vvv){ printf(" close_thxy: max hr_num= %5d (entries/block)\n",hrmax); }
}
/* ---------------------------------------------------------- */
