#
# shellscript, generate povray-picture from daten.def
#  by Joerg.Schulenburg
#
# use:
#  sh def2pov.sh daten.def >a.pov
#  sh def2pov.sh -pov              # => a.ppm a.eps
#
# ChangeLog:
#  v1.1 Aug01 - 3d-def version
#  v1.0 Feb01 - modification of def2fig.sh v1.2
#
# ToDo:
#  - daten.def should have R-vectors for bonds (good for TBC too)
#
if test "$1" == "-pov"; then
 nice -30 povray +FPPM +L/usr/lib/povray31/include +L/usr/lib/povray3/include +W500 +H375 +K0 +Oa.ppm +Ia.pov +A
 # ppmtopgm a.ppm | 
 pnmtops -rle -noturn a.ppm >a.eps
 exit
fi
#
# see /usr/doc/info/gawk.info.gz
# the daten.def file should have additional lines for better drawing:
# #===base=== b1x b1y b2x b2y        - base vektors b1 b2 default=( 1.0 0.0 0.0 1.0 )
#
# if test "$1" == "-pov"; then option="pov"; shift; else option=""; fi
if test -z "$1"; then file=daten.def
                 else file=$1
fi
if test ! -r $file; then echo "file $file not found";exit; fi
#
cat <<END
// povray - raytracer version 2.0
// clock=0..1 for stereoview
#include "colors.inc"
#include "shapes.inc"
#include "chars.inc"
camera {
   location  <10+clock*5, 10,-60> // right,up,away
   look_at   <20,  20,  10> 
}
light_source {< 20, 80, -160> colour White}
background { color White }
union {  // Drei-Bein
 object { char_X scale 0.4 translate <11,0,0> }
 object { char_Y scale 0.4 translate <0,11,0> }
 object { char_Z scale 0.4 translate <0,0,11> }
 cone { < 0, 0, 0>,0.2  < 8,0,0>,0.2 }
 cone { < 0, 0, 0>,0.2  <0, 8,0>,0.2 }
 cone { < 0, 0, 0>,0.2  <0,0, 8>,0.2 }
 cone { < 8, 0, 0>,0.8  <10,0,0>,0.0 }
 cone { < 0, 8, 0>,0.8  <0,10,0>,0.0 }
 cone { < 0, 0, 8>,0.8  <0,0,10>,0.0 }
 pigment { color rgb < 0.7,0.7,0.7 > }
 no_shadow
 translate <-10, -5,-10>
}
#declare Site = 
 sphere { <0, 0, 0>, 3
   pigment { color rgb <0.4,0.4,0.8> }
   finish { ambient 0.2 diffuse 0.9 // refraction 1
   }
   no_shadow
 }
END
awk '\
  function outs( x, y, z, s, c){ # number  c=color s=number
    printf "object { Site scale 1.0 translate <%d,%d,%d> }\n",x,y,z;
  }
  BEGIN { count=0;nr=0;sc=10; ox=0;oy=0;oz=0; rx=0;ry=0;rz=0;
    for( i=0; i<3*3; i++  ) { b[i]=0; e[i]=0; t[i]=0; }
    for( i=0; i<3*3; i+=3 ) { b[i]=1; e[i]=8; t[i]=8; }
    d=2;
  }
  !/^\#/ { count++; nr=count; }
  /^\#/  { nr=0 ; }
  /^\#===base===/  {
    if ( NF<10 ) { d=2 } else { d=3 }
    for( i=0; i<d; i++ ) for( j=0; j<d; j++ ) b[i*3+j]=$(2+i*d+j);
  }
  /^\#===edge===/  {
    if ( NF<10 ) { d=2 } else { d=3 }
    for( i=0; i<d; i++ ) for( j=0; j<d; j++ ) t[i*3+j]=$(2+i*d+j);
    for( i=0; i<3; i++ ) for( j=0; j<3; j++ ) { 
      e[i*3+j]=0;
      for( k=0; k<3; k++ ) {
        e[i*3+j]+=b[k*3+j]*t[i*3+k];
      }
    }
  }
  (nr == 1) { vn=$1;vnw=$2; printf "// NN=%d Nw=%d\n",vn,vnw }
  (nr  > 1) && (nr < vn+2) {\
    if($2<ox) ox=$2; if($2>rx) rx=$2;
    if($3<oy) oy=$3; if($3>ry) ry=$3;
    if($4<oz) oz=$4; if($4>rz) rz=$4;
    sx[$1]=$2; sy[$1]=$3; sz[$1]=$4;
  }
  (nr > vn+1) {\
    w1[$1]=$2;
    w2[$1]=$3;
    wj[$1]=$5;
    if (match($0,"+")) { wb[$1]=1; } else { wb[$1]=0; }
    if ( (sx[$2]-sx[$3])^2 \
       + (sy[$2]-sy[$3])^2 \
       + (sz[$2]-sz[$3])^2 > 4.1 ) { wb[$1]=1; }
  }
  END {
    for ( i=0; i<vn; i++ ) {
      ax[i]=(b[0]*sx[i]+b[3]*sy[i]+b[6]*sz[i]-ox)*sc; 
      ay[i]=(b[1]*sx[i]+b[4]*sy[i]+b[7]*sz[i]-oy)*sc;
      az[i]=(b[2]*sx[i]+b[5]*sy[i]+b[8]*sz[i]-oz)*sc;
      outs(ax[i],ay[i],az[i],i,0);
    }
    for ( i=0; i<vnw; i++ ) {
      s1             ="cone {\n <%d,%d,%d>,0.5\n <%d,%d,%d>,0.5\n pigment { color rgb < 0.25,0.95,0.25 > }\n finish {ambient 0.2 diffuse 0.8 }\n no_shadow }\n"
      if (wj[i]>1) s1="cone {\n <%d,%d,%d>,0.5\n <%d,%d,%d>,0.5\n pigment { color rgb < 0.95,0.25,0.25 > }\n finish {ambient 0.2 diffuse 0.8 }\n no_shadow }\n"
      if (wj[i]>2) s1="cone {\n <%d,%d,%d>,0.5\n <%d,%d,%d>,0.5\n pigment { color rgb < 0.95,0.25,0.95 > }\n finish {ambient 0.2 diffuse 0.8 }\n no_shadow }\n"
      i1=w1[i];
      i2=w2[i];
      printf "// bond %2d %2d - %2d\n",i,i1,i2
      if ( wb[i] != 1 ) printf s1,ax[i1],ay[i1],az[i1],ax[i2],ay[i2],az[i2];
    }
  }
  ' $file
exit

